import steam.webauth as wa
from .err import err
from .err import err_enum
import pathlib
import os
import pickle

class auth:
    s_session_filename = 'session.sb'
    def __init__(self, data_dir:pathlib.Path):
        self.data_dir = data_dir

    def new_session(self, username:str):
        print("This program will NOT save your credential locally")
        print("2FA codes is case SENSITIVE")
        self.username = username
        self.user = wa.WebAuth(username)
        try:
            self.session = self.user.cli_login()
        except:
            raise err(err_enum.LOGIN_FAIL)

        prev_umask = os.umask(0O0077)
        with open(self.get_session_path(), 'wb') as f:
            pickle.dump(self.session.cookies, f)
        os.umask(prev_umask)

    def get_session_path(self):
        return os.path.join(self.data_dir, auth.s_session_filename)



import os
import pathlib
from . import db
from . import err
from .err import err_enum
import datetime
import logging

logger = logging.getLogger('scsd')

class storage:
    s_version_prefix = '.scsd_'

    def __init__(self, location:str, db_:db.db):
        self.location = location
        self.db_ = db_

    def create_game_folder(self, game_name:str, app_id:int):
        db_game_dir = self.db_.get_game_dir(app_id)

        if (db_game_dir is not None):
            if os.path.isdir(os.path.join(self.location, db_game_dir)):
                return
            else:
                dir_name = db_game_dir
        else:
            dir_name = f"{app_id}"

        target_dir = os.path.join(self.location, dir_name)

        if os.path.isdir(target_dir):
            return

        try:
            os.mkdir(target_dir)
        except:
            pass

        if os.path.isdir(target_dir):
            self.db_.set_game_dir(app_id, dir_name)
            return

        # Maybe game_name contains special character fallback to app_id
        dir_name = f"{app_id}__"
        target_dir = os.path.join(self.location, dir_name)

        if os.path.isdir(target_dir):
            return

        try:
            self.db_.set_game_dir(app_id, dir_name)
            os.mkdir(target_dir)
        except:
            raise err.err(err_enum.CANNOT_CREATE_DIRECTORY)

    def get_version_suffix(self, version_num:int):
        if version_num == 0:
            return ""
        else:
            return storage.s_version_prefix + str(version_num)


    # Implicitly create the path if not exist
    def get_filename_location(self,
                              app_id:int,
                              filename:str,
                              file_path:str,
                              version_num:int):
        db_game_dir = self.db_.get_game_dir(app_id)

        version_suffix = self.get_version_suffix(version_num)

        path_to_save = os.path.join(self.location, db_game_dir, file_path)

        if not os.path.isdir(path_to_save):
            os.makedirs(path_to_save)

        return os.path.join(path_to_save, filename + version_suffix)

    def increment_file_version(self,
                               app_id:int,
                               filename:str,
                               file_path:str,
                               current_max_version:int):
        db_game_dir = self.db_.get_game_dir(app_id)

        path_to_save = os.path.join(self.location, db_game_dir, file_path)

        # if db has 5 versions
        # rename
        # version 4 -> 5
        # version 3 -> 4
        # version 2 -> 3
        # version 1 -> 2
        # version 0 -> 1
        for old, new in zip(
                range(current_max_version - 2, -1, -1),
                range(current_max_version - 1, 0, -1)):
            old_version_suffix = self.get_version_suffix(old)
            new_version_suffix = self.get_version_suffix(new)

            os.rename(
                os.path.join(path_to_save, filename + old_version_suffix),
                os.path.join(path_to_save, filename + new_version_suffix))

    # Move current version 0 to 1, 1 to 2, etc
    # Should be done before download file
    def rotate_file(self, app_id:int, filename:str, file_path:str, file_id:int, newest_file_time: datetime.datetime):
        num_of_version = self.db_.update_file_update_time_to_now(file_id, newest_file_time)
        self.increment_file_version(app_id, filename, file_path, num_of_version)

    def remove_outdated(self,
                        app_id:int,
                        filename:str,
                        file_path:str,
                        file_id:int):
        outdated_version_num = self.db_.remove_outdated_file(file_id)

        db_game_dir = self.db_.get_game_dir(app_id)
        path_to_save = os.path.join(self.location, db_game_dir, file_path)

        if outdated_version_num is None:
            return

        for version_num_tup in outdated_version_num:
            version_suffix = self.get_version_suffix(version_num_tup[0])
            logger.info(f"Remove rotated file {filename + version_suffix}")
            try:
                os.remove(os.path.join(path_to_save, filename + version_suffix))
            except:
                e = err.err(err_enum.CANNOT_REMOVE_OUTDATED)
                e.set_additional_info(filename + version_suffix)
                e.log()



import argparse
import os
import pathlib
import logging
from .notifier import notifier
from . import config
from . import ver

class args:
    def __init__(self):
        self.parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

        # Config file has highest priority
        self.parser.add_argument(
            "-f",
            metavar="conf",
            dest="conf",
            type=self.is_file,
            default="",
            help="Path to config file. If given, use the settings and ignore all command arguments"
        )

        self.parser.add_argument(
            "-v",
            "--version",
            dest="version",
            action='version',
            version='%(prog)s-{version}'.format(version=ver.__version__)
        )

        options, remainder = self.parser.parse_known_args()

        # Break if config file is given
        if options.conf is not None:
            return

        self.parser.add_argument(
            "-a",
            metavar="username",
            type=str,
            dest="auth",
            help="Save authentication"
        )

        self.parser.add_argument(
            "-d",
            metavar="save_dir",
            dest="save_dir",
            type=self.is_path,
            required=True,
            help="Directory to save the downloaded saves and db"
        )

        self.parser.add_argument(
            "-r",
            metavar="rotation_count",
            dest="rotation",
            type=int,
            default=config.Defaults['Rotation']['rotation'],
            help="The amount of version for each file to keep"
        )

        self.parser.add_argument(
            "-l",
            metavar="log_level",
            dest="log_level",
            type=int,
            default=config.Defaults['Log']['log_level'],
            help="How detail should the log be"
        )

    def convert_log_level(level:int):
        if (level == 0):
            return logging.ERROR
        elif (level == 1):
            return logging.WARNING
        elif (level == 2):
            return logging.INFO
        else:
            return logging.DEBUG

    def parse(self, raw_args):
        return vars(self.parser.parse_args(raw_args))

    def supported_notifier(self, arg):
        if notifier.is_supported(arg):
            return arg
        else:
            self.parser.error(f"Unsupported notifier '{arg}'")

    def is_file(self, arg):
        if len(arg) == 0:
            return None
        if os.path.isfile(arg):
            return pathlib.Path(arg)
        else:
            self.parser.error(f"The config file '{arg}' does not exist.")

    def is_path(self, arg):
        if os.path.isdir(arg):
            return pathlib.Path(arg)
        else:
            self.parser.error(f"The directory '{arg}' does not exist.")



from . import args
from . import web
from . import db
from . import storage
from . import ver
from .auth import auth
from . import err
from .notifier import notifier
from .config import config
import logging
import sys
import traceback

logger = None

def setup_logger():
    global logger

    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger('scsd')

def parse():
    parsed_args = args.args().parse(sys.argv[1:])

    if parsed_args['conf'] is not None:
        parsed_args = config(parsed_args['conf']).get_conf()
    else:
        parsed_args = config().load_from_arg(parsed_args)

    if 'auth' in parsed_args and \
            parsed_args['auth'] is not None and \
            len(parsed_args['auth']) != 0:
        auth_ = auth(parsed_args['Required']['save_dir'])
        auth_.new_session(parsed_args['auth'])
        exit(0)


    return parsed_args

def should_process_appid(p_target:dict(), p_input:int) -> bool:
    if p_target['mode'] is None:
        return True

    if p_target['mode'] == 'include':
        return p_input in p_target['list']
    elif p_target['mode'] == 'exclude':
        return p_input not in p_target['list']

    return True

def __main__():
    global logger

    notifier_ = None

    try:
        setup_logger()
        parsed_args = parse()

        logger.setLevel(args.args.convert_log_level(parsed_args['Log']['log_level']))
        logger.info(f'scsd-{ver.__version__} started')
        logger.debug(parsed_args)

        notifier_ = notifier.create_instance(
            parsed_args['Notifier']['notifier'],
            **parsed_args['Notifier'])

        summary = main(parsed_args)
    except err.err as e:
        if notifier_:
            notifier_.send(e.get_msg(), False)
        e.log()
        exit(e.num())
    except Exception:
        if notifier_:
            notifier_.send(f"\n```{traceback.format_exc()}```", False)
        print(traceback.format_exc())
        exit(err.err_enum.UNKNOWN_EXCEPTION.value)

    if summary:
        if (len(summary) > 1000):
            summary = summary[0:1000] + "\n...Trunacte\n"
        notifier_.send(summary, True)
    elif parsed_args['Notifier']['notify_if_no_change']:
        notifier_.send("No changes", True)
    exit(0)

def add_new_game(db_, storage_, web_, game, file_infos, summary) -> str:
    db_.add_new_game(game['app_id'], game['name'])
    storage_.create_game_folder(game['name'], game['app_id'])

    file_tuples = [(file_info['filename'], file_info['path'], game['app_id'], file_info['time']) for file_info in file_infos]
    db_.add_new_files(file_tuples)

    summary += f"---\n{game['name']}\n"

    for file_info in file_infos:
        download_game_save(storage_, web_, game, file_info)
        summary += f"↳ {file_info['filename']} (new)\n"

    db_.add_requests_count(len(file_infos) + 1)
    return summary

def download_game_save(storage_, web_, game, file_info):
    logger.info(f"Downloading {file_info['filename']}")
    web_.download_game_save(
        file_info['link'],
        storage_.get_filename_location(game['app_id'],
                                       file_info['filename'],
                                       file_info['path'],
                                       0),
    )

'''
Return true if updated
'''
def update_game(db_, storage_, web_, game, summary) -> tuple:
    logger.info(f"Processing {game['name']}")
    file_infos = web_.get_game_save(game['link'])

    has_update = False

    if (not db_.is_game_exist(game['app_id'])):
        summary = add_new_game(db_, storage_, web_, game, file_infos, summary)
        has_update = True
        return (has_update, summary)

    requests_count = 1
    for file_info in file_infos:
        file_id = db_.get_file_id(game['app_id'], file_info['path'], file_info['filename'])
        if (not db_.is_file_outdated(file_id, file_info['time'])):
            logger.info(f"Ignore {file_info['filename']} (no change)")
            continue

        if not has_update:
            summary += f"---\n{game['name']}\n"

        has_update = True

        storage_.rotate_file(
            game['app_id'],
            file_info['filename'],
            file_info['path'],
            file_id,
            file_info['time'])
        download_game_save(storage_, web_, game, file_info)
        requests_count += 1
        timestamp_sec = file_info['time'].replace(tzinfo=None).isoformat(' ', 'seconds')
        summary += f"↳ {file_info['filename']} ({timestamp_sec})\n"

        storage_.remove_outdated(
            game['app_id'],
            file_info['filename'],
            file_info['path'],
            file_id)

    db_.add_requests_count(requests_count)

    return (has_update, summary)

def main(parsed_args) -> str:

    auth_ = auth(parsed_args['Required']['save_dir'])

    session_pkl = auth_.get_session_path()
    web_ = web.web(session_pkl)

    db_ = db.db(parsed_args['Required']['save_dir'],
                parsed_args['Rotation']['rotation'])
    storage_ = storage.storage(parsed_args['Required']['save_dir'], db_)

    game_list = web_.get_list()

    summary = "\nExecute summary:\n```\n"

    has_update = False
    for game in game_list:
        if db_.is_requests_limit_exceed():
            raise err.err(err.err_enum.REQUESTS_LIMIT_EXCEED)
        if not should_process_appid(parsed_args['Target'], game['app_id']):
            logger.debug(f"Ignoring {game['name']} ({game['app_id']})")
            continue

        game_has_update, summary = \
            update_game(db_, storage_, web_, game, summary)

        has_update = has_update or game_has_update

    if has_update:
        summary += "```"
    else:
        summary = None

    return summary
